import * as Redux from "redux";

export namespace Imdux {
  export interface Action<S, R> {
    dispatch: InferActionDispath<R>;
    query: S;
  }

  export interface Actions {
    [key: string]: Imdux.Action<any, any>;
  }

  export interface Store<T> {
    Dispatch: InferDispatch<T>;
    Query: InferState<T>;
    redux: Redux.Store<any, Redux.AnyAction>;
  }

  export type InferActionDispathFunction<T> = T extends (...arg: infer A) => any
    ? A["length"] extends 2
      ? (payload: A[1]) => void
      : A["length"] extends 1
      ? () => void
      : A["length"] extends 1 | 2
      ? (payload?: A[1]) => void
      : never
    : never;

  export type InferActionDispath<T> = {
    [K in keyof T]: T[K] extends (...arg: any) => any
      ? InferActionDispathFunction<T[K]>
      : InferActionDispath<T[K]>;
  };

  export type InferDispatch<T> = {
    [K in keyof T]: T[K] extends Action<any, infer R>
      ? InferActionDispath<R>
      : unknown;
  };

  export type InferState<T> = {
    [K in keyof T]: T[K] extends Action<infer S, any> ? S : unknown;
  };

  export type Draft<T = any> = (draft: T, payload: any) => T | void;

  export interface Reducers {
    [key: string]: Draft | Reducers;
  }

  export interface CreateActionParams<T> {
    initialState: T;
    reducers: Reducers;
  }

  export interface createStoreOptions {
    devtool: boolean; // false
    payloadNotValidWarn: boolean; // true
    name?: string;
  }
}

export declare function createStore<T extends Imdux.Actions>(
  actions: T,
  options?: Partial<Imdux.createStoreOptions>
): Imdux.Store<T>;

export declare function createAction<S, R>(
  params: Imdux.CreateActionParams<S>
): Imdux.Action<S, R>;
